# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft import serialize
from syft.core.io.address import Address
from syft.grid.messages.group_messages import CreateGroupMessage
from syft.grid.messages.group_messages import CreateGroupResponse
from syft.grid.messages.group_messages import DeleteGroupMessage
from syft.grid.messages.group_messages import DeleteGroupResponse
from syft.grid.messages.group_messages import GetGroupMessage
from syft.grid.messages.group_messages import GetGroupResponse
from syft.grid.messages.group_messages import GetGroupsMessage
from syft.grid.messages.group_messages import GetGroupsResponse
from syft.grid.messages.group_messages import UpdateGroupMessage
from syft.grid.messages.group_messages import UpdateGroupResponse


def test_create_group_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "group-name": "Heart diseases group",
        "members": ["user-id1", "user-id2", "user-id3"],
        "data": [
            {"id": "264632213", "permissions": "read"},
            {"id": "264613232", "permissions": "write"},
            {"id": "896632213", "permissions": "read"},
        ],
    }
    msg = CreateGroupMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_group_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Group Created Successfully!"}
    msg = CreateGroupResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_group_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {"group_id": "f2a6as5d16fasd"}
    msg = DeleteGroupMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_group_response_serde() -> None:
    target = Address(name="Alice")

    content = {"msg": "Group deleted Successfully!"}
    msg = DeleteGroupResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_group_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {
        "group-id": "eqw9e4a5d846",
        "group-name": "Brain diseases group",
        "members": ["user-id1", "user-id2", "user-id3"],
        "data": [
            {"id": "264632213", "permissions": "read"},
            {"id": "264613232", "permissions": "write"},
            {"id": "896632213", "permissions": "read"},
        ],
    }
    msg = UpdateGroupMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_group_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Group updated successfully!"}
    msg = UpdateGroupResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_group_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {"group-id": "eqw9e4a5d846"}
    msg = GetGroupMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_group_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "group-id": "eqw9e4a5d846",
        "group-name": "Heart diseases group",
        "members": ["user-id1", "user-id2", "user-id3"],
        "data": [
            {"id": "264632213", "permissions": "read"},
            {"id": "264613232", "permissions": "write"},
            {"id": "896632213", "permissions": "read"},
        ],
    }

    msg = GetGroupResponse(
        address=target,
        status_code=200,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_groups_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content: Dict[Any, Any] = {}
    msg = GetGroupsMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_groups_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {
        "groups": {
            "626sadaf631": {
                "group-name": "Heart diseases group",
                "members": ["user-id1", "user-id2", "user-id3"],
                "data": [
                    {"id": "264632213", "permissions": "read"},
                    {"id": "264613232", "permissions": "write"},
                    {"id": "896632213", "permissions": "read"},
                ],
            },
            "a84ew64wq6e": {
                "group-name": "Brain diseases group",
                "members": ["user-id5", "user-id7", "user-id9"],
                "data": [
                    {"id": "26463afasd", "permissions": "read"},
                    {"id": "264613dafeqwe", "permissions": "write"},
                    {"id": "896632sdfsf", "permissions": "read"},
                ],
            },
        }
    }

    msg = GetGroupsResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# stdlib

# third party
import torch as th

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.torch.tensor_util import protobuf_tensor_deserializer
from ...lib.torch.tensor_util import protobuf_tensor_serializer
from ...logger import warning
from ...proto.lib.torch.device_pb2 import Device as Device_PB
from ...proto.lib.torch.tensor_pb2 import TensorProto as Tensor_PB

torch_tensor_type = type(th.tensor([1, 2, 3]))


def object2proto(obj: object) -> Tensor_PB:
    proto = Tensor_PB()
    proto.tensor.CopyFrom(protobuf_tensor_serializer(obj))

    proto.requires_grad = getattr(obj, "requires_grad", False)
    proto.device.CopyFrom(
        Device_PB(
            type=obj.device.type,  # type: ignore
            index=obj.device.index,  # type: ignore
        )
    )

    if proto.requires_grad:
        grad = getattr(obj, "grad", None)
        if grad is not None:
            proto.grad.CopyFrom(protobuf_tensor_serializer(grad))

    return proto


def proto2object(proto: Tensor_PB) -> th.Tensor:
    tensor = protobuf_tensor_deserializer(proto.tensor)
    if proto.HasField("grad"):
        tensor.grad = protobuf_tensor_deserializer(proto.grad)

    tensor.requires_grad_(proto.requires_grad)

    if proto.device.type == "cuda" and th.cuda.is_available():
        cuda_index = proto.device.index
        if th.cuda.device_count() < (cuda_index + 1):
            cuda_index = th.cuda.device_count() - 1
            warning(
                f"The requested CUDA index {proto.device.index} is invalid."
                + f"Falling back to GPU index {cuda_index}.",
                print=True,
            )
        return tensor.cuda(cuda_index)

    if proto.device.type == "cuda" and not th.cuda.is_available():
        warning("Cannot find any CUDA devices, falling back to CPU.", print=True)

    return tensor


GenerateWrapper(
    wrapped_type=torch_tensor_type,
    import_path="torch.Tensor",
    protobuf_scheme=Tensor_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
import glob
import json
import os
from pathlib import Path
import re
import shutil
from typing import Any
from typing import Pattern

# third party
import jsonlines
import torch

torch_version = torch.__version__
# --------------------------------------
# add exception and it's handler
# --------------------------------------
exception_pattern_1 = re.compile(
    "no attribute" + "|no NVIDIA driver" + "|Torch not compiled with CUDA enabled"
)


def fix_exception_pattern_1(not_available: list, **kwargs: Any) -> None:
    ele = {
        "lte_version": torch_version,
        "gte_version": torch_version,
        "reason": "no_cpu",
    }
    if ele not in not_available:
        not_available.append(ele)


# errors caused by "self"
exception_pattern_2 = re.compile(
    "convert"
    + "|implement"
    + "|support"
    + "|must be Tensor"
    + "|gradients"
    + "|'CPU' backend"
    + "|xpected"
    + "|can't be cast to the desired output type"
    + "|Can only calculate the mean of"
    + "|input tensor"
    + "|Invalid device, must be xpu device"
    + "|Can only calculate the norm of"
    + "|the base given to float_power_ has dtype"
    + "|sinc_cpu"
)


def fix_exception_pattern_2(
    not_available: list, tensor_type: str, **kwargs: None
) -> None:
    def get_ele_index() -> int:
        keys = {"data_types", "lte_version", "gte_version"}
        i = -1
        for i, ele in enumerate(not_available):
            if (
                set(ele) == keys
                and ele["lte_version"] == torch_version
                and ele["gte_version"] == torch_version
            ):
                return i
        not_available.append(
            {
                "data_types": [],
                "lte_version": torch_version,
                "gte_version": torch_version,
            }
        )
        return i + 1

    i = get_ele_index()
    if tensor_type not in not_available[i]["data_types"]:
        not_available[i]["data_types"].append(tensor_type)


# errors caused by improper armuments
exception_pattern_3 = re.compile("argument")


def fix_exception_pattern_3(not_available: list, inputs: Any, **kwargs: None) -> None:
    def get_ele_index() -> int:
        keys = {"inputs", "lte_version", "gte_version"}
        i = -1
        for i, ele in enumerate(not_available):
            if (
                set(ele) == keys
                and ele["lte_version"] == torch_version
                and ele["gte_version"] == torch_version
            ):
                return i
        not_available.append(
            {
                "inputs": [],
                "lte_version": torch_version,
                "gte_version": torch_version,
            }
        )
        return i + 1

    def exactly_eq(a: Any, b: Any) -> bool:
        """
        `True==1` and `False==0` will return True;
        But we want them be False, so we also check if they are of the same type.
        """
        return type(a) == type(b) and a == b

    i = get_ele_index()
    already_exists = sum([exactly_eq(inputs, _) for _ in not_available[i]["inputs"]])
    if not already_exists:
        not_available[i]["inputs"].append(inputs)


exception_fix = []
exception_fix.append((exception_pattern_1, fix_exception_pattern_1))
exception_fix.append((exception_pattern_2, fix_exception_pattern_2))  # type: ignore
exception_fix.append((exception_pattern_3, fix_exception_pattern_3))  # type: ignore


# ------------------------------
# some helper function
def match_pattern(exception: str, pattern: Pattern) -> int:
    return len(pattern.findall(exception)) > 0


def is_failed_op(line: str, failed_ops: set) -> bool:
    for op in failed_ops:
        if op in line:
            return True
    return False


# --------------------------------
# backup
root_dir = os.path.abspath(Path(os.path.dirname(__file__)) / "..")
p = Path(f"{root_dir}/src/syft/lib/torch/")
shutil.copyfile(p / "allowlist.py", p / "allowlist.py.bak")
shutil.copyfile(
    f"{root_dir}/tests/syft/lib/allowlist_test.json",
    f"{root_dir}/tests/syft/lib/allowlist_test.json.bak",
)


# -------------------------------
# loop:
# [run slow test] -> [comment out allowlist.py] -> [fix allowlist_test.json] -> [run slow test] -> ...
pre_failed_ops = None
same_fail_count = 0
loop_cnt = 0
continue_loop = True
while continue_loop:
    print()
    print("-" * 20 + f"loop {loop_cnt}" + "-" * 20)
    loop_cnt = loop_cnt + 1

    # run slow test
    print("Running slow test ...This may take a while.")
    os.system("pytest -m torch -n auto -p no:benchmark --tb=no")
    print("Slow test done.")
    print()

    # Is errors.jsonl file there?
    err_jsonl = glob.glob(f"{root_dir}/allowlist_test_errors_*.jsonl")
    # if no, all tests pass, stop loop
    if len(err_jsonl) == 0:
        print()
        print("-" * 20)
        print("All tests passed. Auto fix done.")
        os.system(f"git co {p/'allowlist.py'}")
        os.system(f"rm {p/'allowlist.py.bak'}")
        os.system("rm tests/syft/lib/allowlist_test.json.bak")
        break

    # get errors.jsonl file name
    err_jsonl = err_jsonl[0]  # type: ignore

    # get all the operators that failed
    with open(err_jsonl, "r+", encoding="utf8") as f:  # type: ignore
        failed_ops = set([item["input"]["op_name"] for item in jsonlines.Reader(f)])

    print(f"{len(failed_ops)} operators failed:")
    for cnt, op in enumerate(failed_ops):
        print(f"{cnt:>10}:{op:>20}")

    # in case some operators always fail, this loop may go forever
    # we check if it's always the same group of operators that fails for many loops
    if failed_ops == pre_failed_ops:
        same_fail_count = same_fail_count + 1
    else:
        same_fail_count = 0
    if same_fail_count > 5:
        print(f"The same group of operators fail for the last 5 loops. {failed_ops}")
        continue_loop = False
    pre_failed_ops = failed_ops

    # read original content of allowlist.py
    with open(p / "allowlist.py.bak", "r") as f:
        lines = f.readlines()

    # comment all lines but these contain err_ops
    adding_hash = False
    for i, line in enumerate(lines):
        if line.startswith('allowlist["torch'):
            if not is_failed_op(line, failed_ops):
                adding_hash = True
            else:
                adding_hash = False
            # don't comment out `sum` and `backward`, because we will need them when test "grad"
            if "sum" in line or "backward" in line:
                adding_hash = False
        if adding_hash:
            lines[i] = "#" + line

    # overwrite content of allowlist.py
    with open(p / "allowlist.py", "w") as f:
        f.writelines(lines)

    # read allowlist_test.json
    with open(f"{root_dir}/tests/syft/lib/allowlist_test.json", "r") as f:
        allowlist_test = json.load(f)

    # read error.jsonl
    with open(err_jsonl, "r+", encoding="utf8") as f:  # type: ignore
        # process each error line
        for j, err in enumerate(jsonlines.Reader(f)):
            # get error informations
            op_name = err["input"]["op_name"]
            tensor_type = err["input"]["tensor_type"]
            inputs = err["input"]["_args"]
            exception = err["exception"]

            # get the not_available list for this operator
            not_available = []
            if "not_available" in allowlist_test["tests"]["torch.Tensor"][op_name]:
                not_available = allowlist_test["tests"]["torch.Tensor"][op_name][
                    "not_available"
                ]

            # fix this error
            matched = False
            for exception_pattern, fix_method in exception_fix:
                if match_pattern(exception, exception_pattern):
                    fix_method(
                        not_available=not_available,
                        tensor_type=tensor_type,
                        inputs=inputs,
                    )
                    matched = True
            # we suppose to handle all kinds of error, if there is an error we don't handle, raise an exception
            if not matched:
                continue_loop = False
                print(
                    f"A failure can't be handled. Line {j+1} in {err_jsonl}: {exception}"
                )

            # update the not_available
            allowlist_test["tests"]["torch.Tensor"][op_name][
                "not_available"
            ] = not_available

    # update allowlist_test.json
    with open(f"{root_dir}/tests/syft/lib/allowlist_test.json", "w") as f:
        json.dump(allowlist_test, f, indent=2)


# read allowlist_test.json
with open(f"{root_dir}/tests/syft/lib/allowlist_test.json", "r") as f:
    allowlist_test = json.load(f)

# optimize json file
for op, config in allowlist_test["tests"]["torch.Tensor"].items():
    lte_key = "lte_version"
    gte_key = "gte_version"

    # look at not available rules
    if "not_available" in config:
        na_rules = config["not_available"]
        new_rule = None
        for rule in na_rules:
            if lte_key in rule and gte_key in rule:
                rule_lte_version = rule[lte_key]
                rule_gte_version = rule[gte_key]
                if (
                    rule_lte_version == torch_version
                    and rule_gte_version == torch_version
                ):
                    new_rule = rule

        old_rule = None
        match_found = False
        # found a new rule to optimize
        if new_rule is not None:
            for rule in na_rules:
                if new_rule != rule:
                    old_rule_copy = rule.copy()
                    old_rule_copy.pop(lte_key, None)
                    old_rule_copy.pop(gte_key, None)
                    new_rule_copy = new_rule.copy()
                    new_rule_copy.pop(lte_key, None)
                    new_rule_copy.pop(gte_key, None)
                    if old_rule_copy.keys() == new_rule_copy.keys():
                        for k in old_rule_copy.keys():
                            if isinstance(old_rule_copy[k], list) and isinstance(
                                new_rule_copy[k], list
                            ):
                                old_rule_copy[k] = sorted(
                                    old_rule_copy[k], key=lambda x: str(x)
                                )
                                new_rule_copy[k] = sorted(
                                    new_rule_copy[k], key=lambda x: str(x)
                                )

                        # two rules with no version limits and sorted lists match
                        if old_rule_copy == new_rule_copy:
                            match_found = True
                            old_rule = rule.copy()

        # remove the new rule and update the old rule to use current torch_version
        if match_found:
            new_na_rules = []
            for rule in config["not_available"]:
                if rule == new_rule:
                    continue
                if rule == old_rule:
                    rule[lte_key] = torch_version

                new_na_rules.append(rule)
            config["not_available"] = new_na_rules

# update allowlist_test.json
with open(f"{root_dir}/tests/syft/lib/allowlist_test.json", "w") as f:
    json.dump(allowlist_test, f, indent=2)



# stdlib
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# syft relative
from ..core.common.uid import UID
from ..logger import traceback_and_raise
from .attribute import Attribute
from .callable import Callable
from .module import Module


class Globals(Module):
    """The collection of frameworks held in the global namespace"""

    registered_clients: Dict[UID, Any] = {}
    loaded_lib_constructors: Dict[str, CallableT] = {}

    def __init__(
        self,
        client: Optional[Any],
        object_ref: Optional[Union[CallableT, ModuleType]] = None,
        parent: Optional[Attribute] = None,
        path_and_name: Optional[str] = None,
        return_type_name: Optional[str] = None,
    ):
        super().__init__(
            client=client,
            object_ref=object_ref,
            parent=parent,
            path_and_name=path_and_name,
            return_type_name=return_type_name,
        )

    def __call__(
        self,
        path: Union[List[str], str],
        index: int = 0,
        obj_type: Optional[type] = None,
    ) -> Optional[Union[Callable, CallableT]]:

        self.apply_node_changes()

        _path: List[str] = (
            path.split(".") if isinstance(path, str) else path if path else []
        )

        if not _path:
            traceback_and_raise(
                ValueError("Can't execute remote call if path is not specified.")
            )

        return self.attrs[_path[index]](path=_path, index=index + 1, obj_type=obj_type)

    def add_path(
        self,
        path: Union[str, List[str]] = None,  # type:  ignore
        index: int = 0,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        if isinstance(path, str):
            path = path.split(".")

        framework_name = path[index]

        if framework_name not in self.attrs:
            if framework_reference is None:
                traceback_and_raise(
                    Exception(
                        "You must pass in a framework object, the first time you add method \
                        within the framework."
                    )
                )

            self.attrs[framework_name] = Module(
                path_and_name=".".join(path[: index + 1]),
                object_ref=framework_reference,
                return_type_name=return_type_name,
                client=self.client,
                parent=self,
            )

        attr = self.attrs[framework_name]
        attr.add_path(path=path, index=1, return_type_name=return_type_name)

    def register_updates(self, client: Any) -> None:
        # any previously loaded libs need to be applied
        for _, update_ast in self.loaded_lib_constructors.items():
            update_ast(ast_or_client=client)

        # make sure to get any future updates
        self.registered_clients[client.id] = client

    def apply_node_changes(self) -> None:
        pass



# third party
import numpy as np
import torch

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.torch.tensor_util import protobuf_tensor_deserializer
from ...lib.torch.tensor_util import protobuf_tensor_serializer
from ...proto.lib.numpy.array_pb2 import NumpyProto

SUPPORTED_BOOL_TYPES = [np.bool_]
SUPPORTED_INT_TYPES = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

SUPPORTED_FLOAT_TYPES = [
    np.float16,
    np.float32,
    np.float64,
]

SUPPORTED_DTYPES = SUPPORTED_BOOL_TYPES + SUPPORTED_INT_TYPES + SUPPORTED_FLOAT_TYPES

DTYPE_REFACTOR = {
    np.dtype("uint16"): np.int16,
    np.dtype("uint32"): np.int32,
    np.dtype("uint64"): np.int64,
}


def object2proto(obj: np.ndarray) -> NumpyProto:
    original_dtype = obj.dtype
    if original_dtype not in SUPPORTED_DTYPES:
        raise NotImplementedError(f"{original_dtype} is not supported")

    if original_dtype in DTYPE_REFACTOR:
        # store as a signed int, the negative wrap around values convert back to the
        # same original unsigned values on the other side
        obj = obj.astype(DTYPE_REFACTOR[original_dtype])

    tensor = torch.from_numpy(obj).clone()
    tensor_proto = protobuf_tensor_serializer(tensor)
    dtype = original_dtype.name

    return NumpyProto(tensor=tensor_proto, dtype=dtype)


def proto2object(proto: NumpyProto) -> np.ndarray:
    tensor = protobuf_tensor_deserializer(proto.tensor)
    array = tensor.to("cpu").detach().numpy().copy()

    str_dtype = proto.dtype
    original_dtype = np.dtype(str_dtype)

    obj = array.astype(original_dtype)
    return obj


GenerateWrapper(
    wrapped_type=np.ndarray,
    import_path="numpy.ndarray",
    protobuf_scheme=NumpyProto,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
import atexit
from multiprocessing import Process
from multiprocessing import log_to_stderr
from multiprocessing import set_start_method
import socket
import time
from time import sleep
from typing import Callable
from typing import List
from typing import Tuple

# third party
import pytest

# syft relative
from .duet_scenarios_tests import register_duet_scenarios
from .process_test import SyftTestProcess
from .signaling_server_test import run

set_start_method("spawn", force=True)
log_to_stderr()

port = 21000
grid_proc = Process(target=run, args=(port,))
grid_proc.start()


def grid_cleanup() -> None:
    global grid_proc
    grid_proc.terminate()
    grid_proc.join()


atexit.register(grid_cleanup)

registered_tests: List[Tuple[str, Callable, Callable]] = []
register_duet_scenarios(registered_tests)


@pytest.mark.slow
def test_duet() -> None:
    # let the flask server init:
    sleep(3)

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        assert s.connect_ex(("localhost", port)) == 0

    for testcase, do, ds in registered_tests:
        start = time.time()

        do_proc = SyftTestProcess(target=do, args=(port,))
        do_proc.start()

        ds_proc = SyftTestProcess(target=ds, args=(port,))
        ds_proc.start()

        ds_proc.join(30)

        do_proc.terminate()

        if do_proc.exception:
            exception, tb = do_proc.exception
            raise Exception(tb) from exception

        if ds_proc.exception:
            exception, tb = ds_proc.exception
            raise Exception(tb) from exception

        if ds_proc.is_alive():
            ds_proc.terminate()
            raise Exception(f"ds_proc is hanged in {testcase}")

        print(f"test {testcase} passed in {time.time() - start} seconds")



# stdlib
import inspect
import sys
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import List
from typing import Optional
from typing import Union

# syft relative
from .. import ast
from ..ast.callable import Callable
from ..logger import traceback_and_raise


def is_static_method(host_object, attr):  # type: ignore
    value = getattr(host_object, attr)

    # Host object must contain the method resolution order attribute (mro)
    if not hasattr(host_object, "__mro__"):
        return False

    for cls in inspect.getmro(host_object):
        if inspect.isroutine(value):
            if attr in cls.__dict__:
                bound_value = cls.__dict__[attr]
                if isinstance(bound_value, staticmethod):
                    return True
    return False


class Module(ast.attribute.Attribute):
    """A module which contains other modules or callables."""

    def __init__(
        self,
        client: Optional[Any],
        parent: Optional[ast.attribute.Attribute] = None,
        path_and_name: Optional[str] = None,
        object_ref: Optional[Union[CallableT, ModuleType]] = None,
        return_type_name: Optional[str] = None,
    ):
        """
        Args:
            client: The client for which all computation is being executed.
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The given action's return type name, with its full path, in string format.
        """
        super().__init__(
            path_and_name=path_and_name,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )

        if object_ref is None and self.name:
            try:
                self.object_ref = sys.modules[path_and_name if path_and_name else ""]
            except Exception:
                self.object_ref = getattr(self.parent.object_ref, self.name)

    def add_attr(
        self,
        attr_name: str,
        attr: Optional[Union[Callable, CallableT]],
        is_static: bool = False,
    ) -> None:
        """
        Add an attribute to the current module.

        Args:
            attr_name: The name of the attribute, e.g. `List` of the path `syft.lib.python.List`.
            attr: The attribute object.
            is_static: The actual Python object for which the computation is being made.
        """
        self.__setattr__(attr_name, attr)

        if is_static is True:
            traceback_and_raise(
                ValueError("Static methods shouldn't be added to an object.")
            )

        if attr is None:
            traceback_and_raise(ValueError("An attribute reference has to be passed."))

        # If `add_attr` is called directly, we need to cache the path as well
        attr_ref = getattr(attr, "object_ref", None)
        path = getattr(attr, "path_and_name", None)
        if attr_ref not in self.lookup_cache and path is not None:
            self.lookup_cache[attr_ref] = path

        self.attrs[attr_name] = attr  # type: ignore

    def __call__(
        self,
        path: Union[List[str], str],
        index: int = 0,
        obj_type: Optional[type] = None,
    ) -> Optional[Union[Callable, CallableT]]:

        self.apply_node_changes()

        if obj_type is not None:
            if obj_type in self.lookup_cache:
                path = self.lookup_cache[obj_type]

        _path: List[str] = (
            path.split(".") if isinstance(path, str) else path if path else []
        )

        if not _path:
            traceback_and_raise(
                ValueError("Can't execute remote call if path is not specified.")
            )

        resolved = self.attrs[_path[index]](
            path=_path, index=index + 1, obj_type=obj_type
        )

        return resolved

    def __repr__(self) -> str:
        out = "Module:\n"
        for name, module in self.attrs.items():
            out += "\t." + name + " -> " + str(module).replace("\t.", "\t\t.") + "\n"

        return out

    def add_path(
        self,
        path: List[str],
        index: int = 0,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        if index >= len(path):
            return

        if path[index] not in self.attrs:
            attr_ref = getattr(self.object_ref, path[index])

            if inspect.ismodule(attr_ref):
                self.add_attr(
                    attr_name=path[index],
                    attr=ast.module.Module(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        parent=self,
                    ),
                )
            elif inspect.isclass(attr_ref):
                klass = ast.klass.Class(
                    path_and_name=".".join(path[: index + 1]),
                    object_ref=attr_ref,
                    return_type_name=return_type_name,
                    client=self.client,
                    parent=self,
                )
                self.add_attr(
                    attr_name=path[index],
                    attr=klass,
                )
            elif inspect.isfunction(attr_ref) or inspect.isbuiltin(attr_ref):
                is_static = is_static_method(self.object_ref, path[index])  # type: ignore

                self.add_attr(
                    attr_name=path[index],
                    attr=ast.callable.Callable(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        is_static=is_static,
                        parent=self,
                    ),
                )
            elif inspect.isdatadescriptor(attr_ref):
                self.add_attr(
                    attr_name=path[index],
                    attr=ast.property.Property(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        parent=self,
                    ),
                )
            elif index == len(path) - 1:
                static_attribute = ast.static_attr.StaticAttribute(
                    path_and_name=".".join(path[: index + 1]),
                    return_type_name=return_type_name,
                    client=self.client,
                    parent=self,
                )
                setattr(self, path[index], static_attribute)
                self.attrs[path[index]] = static_attribute
                return

        attr = self.attrs[path[index]]
        attr_ref = getattr(self.object_ref, path[index], None)
        if attr_ref is not None and attr_ref not in self.lookup_cache:
            self.lookup_cache[attr_ref] = path

        attr.add_path(path=path, index=index + 1, return_type_name=return_type_name)

    def __getattribute__(self, item: str) -> Any:
        target_object = super().__getattribute__(item)
        if isinstance(target_object, ast.static_attr.StaticAttribute):
            return target_object.get_remote_value()
        return target_object

    def __setattr__(self, key: str, value: Any) -> None:
        if hasattr(super(), "attrs"):
            attrs = super().__getattribute__("attrs")
            if key in attrs:
                target_object = self.attrs[key]
                if isinstance(target_object, ast.static_attr.StaticAttribute):
                    return target_object.set_remote_value(value)

        return super().__setattr__(key, value)

    def fetch_live_object(self) -> object:
        try:
            return sys.modules[self.path_and_name if self.path_and_name else ""]
        except Exception:
            return getattr(self.parent.object_ref, self.name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/zksk/nizk.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.petlib import bn_pb2 as proto_dot_lib_dot_petlib_dot_bn__pb2
from syft.proto.lib.python import tuple_pb2 as proto_dot_lib_dot_python_dot_tuple__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/zksk/nizk.proto",
    package="syft.lib.zksk",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x19proto/lib/zksk/nizk.proto\x12\rsyft.lib.zksk\x1a\x19proto/lib/petlib/bn.proto\x1a\x1cproto/lib/python/tuple.proto"l\n\x04NIZK\x12&\n\tchallenge\x18\x01 \x01(\x0b\x32\x13.syft.lib.petlib.Bn\x12)\n\tresponses\x18\x02 \x01(\x0b\x32\x16.syft.lib.python.Tuple\x12\x11\n\tstmt_hash\x18\x03 \x01(\x0c\x62\x06proto3',
    dependencies=[
        proto_dot_lib_dot_petlib_dot_bn__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_tuple__pb2.DESCRIPTOR,
    ],
)


_NIZK = _descriptor.Descriptor(
    name="NIZK",
    full_name="syft.lib.zksk.NIZK",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="challenge",
            full_name="syft.lib.zksk.NIZK.challenge",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="responses",
            full_name="syft.lib.zksk.NIZK.responses",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stmt_hash",
            full_name="syft.lib.zksk.NIZK.stmt_hash",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=101,
    serialized_end=209,
)

_NIZK.fields_by_name[
    "challenge"
].message_type = proto_dot_lib_dot_petlib_dot_bn__pb2._BN
_NIZK.fields_by_name[
    "responses"
].message_type = proto_dot_lib_dot_python_dot_tuple__pb2._TUPLE
DESCRIPTOR.message_types_by_name["NIZK"] = _NIZK
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NIZK = _reflection.GeneratedProtocolMessageType(
    "NIZK",
    (_message.Message,),
    {
        "DESCRIPTOR": _NIZK,
        "__module__": "proto.lib.zksk.nizk_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.zksk.NIZK)
    },
)
_sym_db.RegisterMessage(NIZK)


# @@protoc_insertion_point(module_scope)



"""In this test suite, we evaluate the Address class. For more info
on the Address class and its purpose, please see the documentation
in the class itself.

Table of Contents:
    - INITIALIZATION: tests for various ways Address can/can't be initialized
    - CLASS METHODS: tests for the use of Address's class methods
    - SERDE: test for serialization and deserialization of Address.

"""


# stdlib
from itertools import combinations
import uuid

# third party
import pytest

# syft absolute
import syft as sy
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.io.location.specific import SpecificLocation

ARGUMENTS = ["vm", "device", "domain", "network"]

# --------------------- INITIALIZATION ---------------------


def test_init_without_arguments() -> None:
    """Test that Address have all attributes as None if none are given"""

    # init works without arguments
    addr = Address()

    assert addr.network is None
    assert addr.domain is None
    assert addr.device is None
    assert addr.vm is None

    with pytest.raises(Exception):
        assert addr.target_id is None


def _gen_address_kwargs() -> list:
    """
    Helper method to generate pre-ordered arguments for initializing an Address instance.
    There are at least 3 arguments, all taken from 'vm', 'device', 'domain', 'network'.
    """
    # the order matches the _gen_icons below
    all_combos = []
    for combination in combinations(ARGUMENTS, 3):
        all_combos.append(list(combination))
    all_combos.append(ARGUMENTS)
    return [{key: SpecificLocation(id=UID()) for key in combo} for combo in all_combos]


def _gen_icons() -> list:
    """Helper method to return an pre-ordered list of icons."""
    return [
        "💠 [🍰📱🏰]",
        "💠 [🍰📱🔗]",
        "💠 [🍰🏰🔗]",
        "💠 [📱🏰🔗]",
        "💠 [🍰📱🏰🔗]",
    ]


def _gen_address_kwargs_and_expected_values() -> list:
    """
    Helper method to generate kwargs for initializing Address as well as
    the expected values thereof.
    """
    address_kwargs = _gen_address_kwargs()
    expected_value_dict = [
        {key: kwargs.get(key, None) for key in ARGUMENTS} for kwargs in address_kwargs
    ]
    return list(zip(address_kwargs, expected_value_dict))


@pytest.mark.parametrize(
    "address_kwargs, expected_values", _gen_address_kwargs_and_expected_values()
)
def test_init_with_specific_id(address_kwargs: dict, expected_values: dict) -> None:
    """Test that Address will use the SpecificLocation you pass into the constructor"""
    address = Address(**address_kwargs)

    assert address.network is expected_values["network"]
    assert address.domain is expected_values["domain"]
    assert address.device is expected_values["device"]
    assert address.vm is expected_values["vm"]


# --------------------- CLASS METHODS ---------------------


def test_compare() -> None:
    """Tests whether two address objects are the same. This functionality
    is likely to get used a lot when nodes are determining whether a message
    is for them or not."""

    x = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )

    y = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )

    z = Address(network=x.network, domain=x.domain, device=x.device, vm=x.vm)

    assert x != y
    assert x == z
    assert y != z


def test_to_string() -> None:
    """Tests that SpecificLocation generates an intuitive string."""

    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))

    obj = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    str_out = (
        "<Address - Network:<SpecificLocation:..a1514>, Domain:<Specific"
        "Location:..a1514>  Device:<SpecificLocation:..a1514>, VM:<Specific"
        "Location:..a1514>"
    )

    assert str(obj) == str_out
    assert obj.__repr__() == str_out


def test_target_emoji_method() -> None:
    """Unit test for Address.target_emoji method"""
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))

    address = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )
    assert address.target_emoji() == "@<UID:🙍🛖>"


# --------------------- PROPERTY METHODS ---------------------


@pytest.mark.parametrize(
    "address_kwargs, expected_icon",
    list(
        zip(
            _gen_address_kwargs(),
            _gen_icons(),
        )
    ),
)
def test_icon_property_method(address_kwargs: dict, expected_icon: str) -> None:
    """Unit tests for Address.icon property method"""
    address = Address(**address_kwargs)
    assert address.icon == expected_icon


@pytest.mark.parametrize(
    "address_kwargs, expected_icon",
    list(
        zip(
            _gen_address_kwargs(),
            _gen_icons(),
        )
    ),
)
def test_pprint_property_method(address_kwargs: dict, expected_icon: str) -> None:
    """Unit tests for Address.pprint property method"""
    named_address = Address(name="Sneaky Nahua", **address_kwargs)
    assert named_address.pprint == expected_icon + " Sneaky Nahua (Address)"

    unnamed_address = Address(**address_kwargs)
    assert expected_icon in unnamed_address.pprint
    assert "(Address)" in unnamed_address.pprint


def test_address_property_method() -> None:
    """Unit tests for Address.address property method"""
    address = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )

    returned_address = address.address
    assert isinstance(returned_address, Address)
    assert returned_address.network == address.network
    assert returned_address.domain == address.domain
    assert returned_address.device == address.device
    assert returned_address.vm == address.vm


def test_network_getter_and_setter() -> None:
    """Unit tests for Address.network getter and setter"""
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    # Test getter
    network = SpecificLocation(id=an_id)
    address = Address(
        network=network,
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )
    assert address.network == network

    # Test setter
    new_network = SpecificLocation(id=an_id)
    address.network = new_network
    assert address.network == new_network


def test_network_id_property_method() -> None:
    """Unit test for Address.network_id method"""
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    # Test getter
    address_with_network = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )
    address_without_network = Address(
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    assert address_with_network.network_id == an_id
    assert address_without_network.network_id is None


def test_domain_and_domain_id_property_methods() -> None:
    """Unit test for Address.domain and Address.domain_id methods"""
    # Test getter
    domain = SpecificLocation(id=UID())
    address_with_domain = Address(
        network=SpecificLocation(id=UID()),
        domain=domain,
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )
    # Test domain getter
    assert address_with_domain.domain == domain

    # Test domain setter
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    new_domain = SpecificLocation(id=an_id)
    address_with_domain.domain = new_domain
    assert address_with_domain.domain == new_domain

    # Test domain_id getter
    address_without_domain = Address(
        network=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )
    assert address_with_domain.domain_id == an_id
    assert address_without_domain.domain_id is None


def test_device_and_device_id_property_methods() -> None:
    """Unit test for Address.device and Address.device_id methods"""
    # Test getter
    device = SpecificLocation(id=UID())
    address_with_device = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=device,
        vm=SpecificLocation(id=UID()),
    )
    # Test device getter
    assert address_with_device.device == device

    # Test device setter
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    new_device = SpecificLocation(id=an_id)
    address_with_device.device = new_device
    assert address_with_device.device == new_device

    # Test domain_id getter
    address_without_device = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )
    assert address_with_device.device_id == an_id
    assert address_without_device.device_id is None


def test_vm_and_vm_id_property_methods() -> None:
    """Unit test for Address.vm and Address.vm_id methods"""
    # Test getter
    vm = SpecificLocation(id=UID())
    address_with_vm = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=vm,
    )
    # Test device getter
    assert address_with_vm.vm == vm

    # Test device setter
    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    new_vm = SpecificLocation(id=an_id)
    address_with_vm.vm = new_vm
    assert address_with_vm.vm == new_vm

    # Test domain_id getter
    address_without_vm = Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
    )
    assert address_with_vm.vm_id == an_id
    assert address_without_vm.vm_id is None


def test_target_id_property_method_with_a_return() -> None:
    """Unit test for Address.target_id method"""
    network = SpecificLocation(id=UID())
    domain = SpecificLocation(id=UID())
    device = SpecificLocation(id=UID())
    vm = SpecificLocation(id=UID())
    address = Address(
        network=network,
        domain=domain,
        device=device,
        vm=vm,
    )
    assert address.target_id == vm
    address.vm = None
    assert address.target_id == device
    address.device = None
    assert address.target_id == domain
    address.domain = None
    assert address.target_id == network


# --------------------- SERDE ---------------------


def test_default_serialization_and_deserialization() -> None:
    """Tests that default Address serde works as expected - to Protobuf"""

    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))

    obj = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    blob = sy.serialize(obj, to_proto=True)

    assert sy.serialize(obj) == blob
    assert obj == sy.deserialize(blob=blob)


def test_partial_serialization_and_deserialization() -> None:
    """Test that addresses with only some attributes serialize and deserialize correctly/"""

    an_id = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))

    obj = Address(  # network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    assert obj == sy.deserialize(blob=sy.serialize(obj))

    obj = Address(
        network=SpecificLocation(id=an_id),
        # domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    blob = sy.serialize(obj, to_proto=True)
    assert obj == sy.deserialize(blob=blob)

    obj = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        # device=SpecificLocation(id=an_id),
        vm=SpecificLocation(id=an_id),
    )

    blob = sy.serialize(obj, to_proto=True)
    assert obj == sy.deserialize(blob=blob)

    obj = Address(
        network=SpecificLocation(id=an_id),
        domain=SpecificLocation(id=an_id),
        device=SpecificLocation(id=an_id),
        # vm=SpecificLocation(id=an_id)
    )

    blob = sy.serialize(obj, to_proto=True)
    assert obj == sy.deserialize(blob=blob)


def test_proto_serialization() -> None:
    """Tests that default Address serialization works as expected - to Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    loc = SpecificLocation(id=uid, name="Test Location")
    obj = Address(
        name="Test Address",
        network=loc,
        domain=loc,
        device=loc,
        vm=loc,
    )

    blob = Address.get_protobuf_schema()(
        name="Test Address",
        has_network=True,
        has_domain=True,
        has_device=True,
        has_vm=True,
        network=sy.serialize(loc),
        domain=sy.serialize(loc),
        device=sy.serialize(loc),
        vm=sy.serialize(loc),
    )

    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob


def test_proto_deserialization() -> None:
    """Tests that default Address deserialization works as expected - from Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    loc = SpecificLocation(id=uid, name="Test Location")

    obj = Address(
        network=loc,
        domain=loc,
        device=loc,
        vm=loc,
    )

    blob = Address.get_protobuf_schema()(
        has_network=True,
        has_domain=True,
        has_device=True,
        has_vm=True,
        network=sy.serialize(loc),
        domain=sy.serialize(loc),
        device=sy.serialize(loc),
        vm=sy.serialize(loc),
    )

    obj2 = sy.deserialize(blob=blob, from_proto=True)
    assert obj == obj2



# stdlib
from typing import Dict
from typing import Union

# TODO: Refactor out all the test data.
# Issue: https://github.com/OpenMined/PySyft/issues/5325

allowlist: Dict[str, Union[str, Dict[str, str]]] = {}  # (path: str, return_type:type)

allowlist["torchvision.__version__"] = "syft.lib.python.String"
# MNIST
allowlist["torchvision.transforms.Compose"] = "torchvision.transforms.Compose"
# allowlist["torchvision.transforms.Compose.__iter__"] = "torchvision.transforms.ToTensor"
# TODO: Compose.transforms property only exists on the object not on the class?
# allowlist["torchvision.transforms.Compose.transforms"] = "syft.lib.python.List"
allowlist["torchvision.transforms.ToTensor"] = "torchvision.transforms.ToTensor"
allowlist["torchvision.transforms.Normalize"] = "torchvision.transforms.Normalize"
# TODO: Normalize properties only exists on the object not on the class?
# allowlist["torchvision.transforms.Normalize.inplace"] = "syft.lib.python.Bool"
# TODO: mean and std are actually tuples
# allowlist["torchvision.transforms.Normalize.mean"] = "syft.lib.python.List"
# allowlist["torchvision.transforms.Normalize.std"] = "syft.lib.python.List"

# Datasets

allowlist["torchvision.datasets.MNIST"] = {
    "return_type": "torchvision.datasets.MNIST",
    "test_parameters": "('../data', download=False,)",
}


allowlist["torchvision.datasets.MNIST.__len__"] = "syft.lib.python.Int"


allowlist["torchvision.datasets.CelebA"] = {
    "return_type": "torchvision.datasets.CelebA",
    "test_parameters": "('../data')",
}
allowlist["torchvision.datasets.CelebA.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CIFAR10"] = {
    "return_type": "torchvision.datasets.CIFAR10",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.CIFAR10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CIFAR100"] = {
    "return_type": "torchvision.datasets.CIFAR100",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.CIFAR10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Cityscapes"] = {
    "return_type": "torchvision.datasets.Cityscapes",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.Cityscapes.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CocoCaptions"] = {
    "return_type": "torchvision.datasets.CocoCaptions",
    "test_parameters": "('../data','../data/captions.txt')",
}
allowlist["torchvision.datasets.CocoCaptions.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CocoDetection"] = {
    "return_type": "torchvision.datasets.CocoDetection",
    "test_parameters": "('../data', '../data/captions.txt')",
}
allowlist["torchvision.datasets.CocoDetection.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.EMNIST"] = {
    "return_type": "torchvision.datasets.EMNIST",
    "test_parameters": "('../data',split = \"mnist\")",
}
allowlist["torchvision.datasets.EMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.FakeData"] = {
    "return_type": "torchvision.datasets.FakeData",
    "test_parameters": "('../data', )",
}
allowlist["torchvision.datasets.FakeData.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.FashionMNIST"] = {
    "return_type": "torchvision.datasets.FashionMNIST",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.FashionMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Flickr8k"] = {
    "return_type": "torchvision.datasets.Flickr8k",
    "test_parameters": "('../data', '../data/annfile.txt')",
}
allowlist["torchvision.datasets.Flickr8k.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Flickr30k"] = {
    "return_type": "torchvision.datasets.Flickr30k",
    "test_parameters": "('../data', '../data/annfile.txt')",
}
allowlist["torchvision.datasets.Flickr30k.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.HMDB51"] = {
    "return_type": "torchvision.datasets.HMDB51",
    "test_parameters": "('../data', '../data/annfile.txt', 20,  )",
}
allowlist["torchvision.datasets.HMDB51.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.ImageNet"] = {
    "return_type": "torchvision.datasets.ImageNet",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.ImageNet.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Kinetics400"] = {
    "return_type": "torchvision.datasets.Kinetics400",
    "test_parameters": "('../data', 20)",
}
allowlist["torchvision.datasets.Kinetics400.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.KMNIST"] = {
    "return_type": "torchvision.datasets.KMNIST",
    "test_parameters": "('../data', )",
}
allowlist["torchvision.datasets.KMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.LSUN"] = {
    "return_type": "torchvision.datasets.LSUN",
    "test_parameters": "('../data', )",
}
allowlist["torchvision.datasets.LSUN.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Omniglot"] = {
    "return_type": "torchvision.datasets.Omniglot",
    "min_version": "0.8.0",
    "test_parameters": "('../data', )",
}
allowlist["torchvision.datasets.Omniglot.__len__"] = {
    "return_type": "syft.lib.python.Int",
    "min_version": "0.8.0",
}

allowlist["torchvision.datasets.PhotoTour"] = {
    "return_type": "torchvision.datasets.PhotoTour",
    "test_parameters": "('../data', name = 'data')",
}
allowlist["torchvision.datasets.PhotoTour.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Places365"] = {
    "return_type": "torchvision.datasets.Places365",
    "min_version": "0.8.0",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.Places365.__len__"] = {
    "return_type": "syft.lib.python.Int",
    "min_version": "0.8.0",
}

allowlist["torchvision.datasets.QMNIST"] = {
    "return_type": "torchvision.datasets.QMNIST",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.QMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SBDataset"] = {
    "return_type": "torchvision.datasets.SBDataset",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.SBDataset.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SBU"] = {
    "return_type": "torchvision.datasets.SBU",
    "test_parameters": "('../data', download = False)",
}
allowlist["torchvision.datasets.SBU.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.STL10"] = {
    "return_type": "torchvision.datasets.STL10",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.STL10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SVHN"] = {
    "return_type": "torchvision.datasets.SVHN",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.SVHN.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.UCF101"] = {
    "return_type": "torchvision.datasets.UCF101",
    "test_parameters": "('../data', frames_per_clip = 20, annotation_path = '../data/annfile.txt')",
}
allowlist["torchvision.datasets.UCF101.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.USPS"] = {
    "return_type": "torchvision.datasets.USPS",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.USPS.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.VOCSegmentation"] = {
    "return_type": "torchvision.datasets.VOCSegmentation",
    "test_parameters": "('../data',)",
}
allowlist["torchvision.datasets.VOCSegmentation.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.VOCDetection"] = {
    "return_type": "torchvision.datasets.VOCDetection",
    "test_parameters": "('../data',)",
}
allowlist[
    "torchvision.datasets.VOCDetection.__len__"
] = "torchvision.datasets.VOCDetection"

# Transforms
allowlist["torchvision.transforms.CenterCrop"] = {
    "return_type": "torchvision.transforms.CenterCrop",
    "test_parameters": "(10)",
}
allowlist["torchvision.transforms.ColorJitter"] = {
    "return_type": "torchvision.transforms.ColorJitter",
    "test_parameters": "(brightness=0, contrast=0, saturation=0, hue=0)",
}

# This is an interesting case, for some versions p = 0.2 is needed, for others its not needed

allowlist["torchvision.transforms.FiveCrop"] = {
    "return_type": "torchvision.transforms.FiveCrop",
    #    "test_parameters": "(size = 10, p = 0.2)",
}


allowlist["torchvision.transforms.Grayscale"] = {
    "return_type": "torchvision.transforms.Grayscale",
    "test_parameters": "(num_output_channels=1)",
}
allowlist["torchvision.transforms.Pad"] = {
    "return_type": "torchvision.transforms.Pad",
    "test_parameters": "(2, fill=0, padding_mode='constant')",
}
allowlist["torchvision.transforms.RandomAffine"] = {
    "return_type": "torchvision.transforms.RandomAffine",
    "test_parameters": "(degrees = 2)",
}

# transforms error
allowlist["torchvision.transforms.RandomApply"] = {
    "return_type": "torchvision.transforms.RandomApply",
    # "test_parameters": "(torchvision.transforms.CenterCrop(10))",
}

allowlist["torchvision.transforms.RandomCrop"] = {
    "return_type": "torchvision.transforms.RandomCrop",
    "test_parameters": "(size = 10)",
}
allowlist["torchvision.transforms.RandomGrayscale"] = {
    "return_type": "torchvision.transforms.RandomGrayscale",
    "test_parameters": "(p=0.1)",
}

allowlist["torchvision.transforms.RandomHorizontalFlip"] = {
    "return_type": "torchvision.transforms.RandomHorizontalFlip",
    "test_parameters": "(p=0.1)",
}
allowlist["torchvision.transforms.RandomPerspective"] = {
    "return_type": "torchvision.transforms.RandomPerspective",
    "test_parameters": "(distortion_scale=0.5, p=0.5)",
}

allowlist["torchvision.transforms.RandomResizedCrop"] = {
    "return_type": "torchvision.transforms.RandomResizedCrop",
    "test_parameters": "(10, scale=(0.08, 1.0), ratio=(0.75, 1.25))",
}
allowlist["torchvision.transforms.RandomRotation"] = {
    "return_type": "torchvision.transforms.RandomRotation",
    "test_parameters": "(degrees = 2)",
}
allowlist["torchvision.transforms.RandomSizedCrop"] = {
    "return_type": "torchvision.transforms.RandomSizedCrop",
    "test_parameters": "(10)",
}
allowlist["torchvision.transforms.RandomVerticalFlip"] = {
    "return_type": "torchvision.transforms.RandomVerticalFlip",
    "test_parameters": "(p=0.5)",
}
allowlist["torchvision.transforms.Resize"] = {
    "return_type": "torchvision.transforms.Resize",
    "test_parameters": "(size = 15)",
}
allowlist["torchvision.transforms.Scale"] = {
    "return_type": "torchvision.transforms.Scale",
    "test_parameters": "(10)",
}
allowlist["torchvision.transforms.TenCrop"] = {
    "return_type": "torchvision.transforms.TenCrop",
    "test_parameters": "(10)",
}
allowlist["torchvision.transforms.GaussianBlur"] = {
    "return_type": "torchvision.transforms.GaussianBlur",
    "min_version": "0.8.0",
    "test_parameters": "(kernel_size = 3)",
}

allowlist["torchvision.transforms.RandomChoice"] = {
    "return_type": "torchvision.transforms.RandomChoice",
}
allowlist["torchvision.transforms.RandomOrder"] = {
    "return_type": "torchvision.transforms.RandomOrder",
}

allowlist[
    "torchvision.transforms.LinearTransformation"
] = "torchvision.transforms.LinearTransformation"
allowlist[
    "torchvision.transforms.RandomErasing"
] = "torchvision.transforms.RandomErasing"
allowlist["torchvision.transforms.ConvertImageDtype"] = {
    "return_type": "torchvision.transforms.ConvertImageDtype",
    "min_version": "0.8.0",
}
allowlist["torchvision.transforms.ToPILImage"] = "torchvision.transforms.ToPILImage"
allowlist["torchvision.transforms.Lambda"] = "torchvision.transforms.Lambda"

# Functional Transformers
allowlist["torchvision.transforms.functional.adjust_brightness"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 0.5)",
}
allowlist["torchvision.transforms.functional.adjust_contrast"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 0.5)",
}
allowlist["torchvision.transforms.functional.adjust_gamma"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 1, 0.5)",
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.adjust_hue"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 0)"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.adjust_saturation"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 0.5)",
}
allowlist["torchvision.transforms.functional.adjust_sharpness"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens, 0.5)",
}
allowlist["torchvision.transforms.functional.affine"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens,0.2,[1,2],0.2,[1,2])",
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.autocontrast"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens)",
}
allowlist["torchvision.transforms.functional.center_crop"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 10)",
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.convert_image_dtype"] = {
    "return_type": "torch.Tensor",
}
allowlist["torchvision.transforms.functional.crop"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 10 , 20, 30, 40)",
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.equalize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens)",
}
allowlist["torchvision.transforms.functional.erase"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, 10, 20, 30, 40, 250)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.five_crop"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, 10)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.gaussian_blur"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens, 3)",
}
allowlist["torchvision.transforms.functional.hflip"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.invert"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens)",
}
allowlist["torchvision.transforms.functional.normalize"] = {
    "return_type": "torch.Tensor",
    # "test_parameters": "(tens, [0.5, 0.5, 0.5], [1, 1, 1]).unsqueeze(0)",
    # currently commenting because of test issues with hier versions
    # (//) works for 1.6.0 and / works for higher version :(
}
allowlist["torchvision.transforms.functional.pad"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, 10)",
    "min_version": "0.8.0"
    # # Torch 1.6 expects input to be PIL image, so commenting this currently
}
allowlist["torchvision.transforms.functional.perspective"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, [[10,20],[20,30],[30, 40],[40,50]], [[20,30],[30,40],[40, 50],[50,60]])",
    "min_version": "0.8.0"
    # # Torch 1.6 expects input to be PIL image, so commenting this currently
}

# Converts PIL to tensor, currently not supported
# allowlist["torchvision.transforms.functional.pil_to_tensor"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.posterize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens, 2)",
}
allowlist["torchvision.transforms.functional.resize"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, 10)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.resized_crop"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, 10, 15, 20, 25, 30)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.rgb_to_grayscale"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
    "test_parameters": "(tens)",
}
allowlist["torchvision.transforms.functional.rotate"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, angle = 10)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.solarize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
    "test_parameters": "(tens, threshold = 0.5)",
}
allowlist["torchvision.transforms.functional.ten_crop"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens, size = 10)",
    "min_version": "0.8.0"
    # Torch 1.6 expects input to be PIL image, so minimum version as 0.7 (Torch 1.7.0)
}
allowlist["torchvision.transforms.functional.to_grayscale"] = {
    "return_type": "PIL.Image.Image"
}
allowlist["torchvision.transforms.functional.to_pil_image"] = {
    "return_type": "PIL.Image.Image",
}
allowlist["torchvision.transforms.functional.to_tensor"] = {
    "return_type": "torch.Tensor",
}
allowlist["torchvision.transforms.functional.vflip"] = {
    "return_type": "torch.Tensor",
    "test_parameters": "(tens)",
}



# stdlib
from typing import Any

# syft relative
from ...core.common import UID
from ...core.common.serde.serializable import Serializable


class PyPrimitive(Serializable):
    def __init__(self) -> None:
        self._id: UID

    def upcast(self) -> Any:
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/action.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.node.common.action import (
    garbage_collect_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_garbage__collect__object__pb2,
)
from syft.proto.core.node.common.action import (
    get_enum_attribute_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__enum__attribute__pb2,
)
from syft.proto.core.node.common.action import (
    get_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__object__pb2,
)
from syft.proto.core.node.common.action import (
    get_set_property_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__property__pb2,
)
from syft.proto.core.node.common.action import (
    get_set_static_attribute_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__static__attribute__pb2,
)
from syft.proto.core.node.common.action import (
    run_class_method_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_run__class__method__pb2,
)
from syft.proto.core.node.common.action import (
    run_function_or_constructor_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_run__function__or__constructor__pb2,
)
from syft.proto.core.node.common.action import (
    save_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_save__object__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/action/action.proto",
    package="syft.core.node.common.action",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n*proto/core/node/common/action/action.proto\x12\x1csyft.core.node.common.action\x1a.proto/core/node/common/action/get_object.proto\x1a?proto/core/node/common/action/run_function_or_constructor.proto\x1a\x34proto/core/node/common/action/run_class_method.proto\x1a:proto/core/node/common/action/garbage_collect_object.proto\x1a\x36proto/core/node/common/action/get_enum_attribute.proto\x1a\x34proto/core/node/common/action/get_set_property.proto\x1a<proto/core/node/common/action/get_set_static_attribute.proto\x1a/proto/core/node/common/action/save_object.proto"\xfd\x05\n\x06\x41\x63tion\x12\x10\n\x08obj_type\x18\x01 \x01(\t\x12J\n\x11get_object_action\x18\x02 \x01(\x0b\x32-.syft.core.node.common.action.GetObjectActionH\x00\x12j\n"run_function_or_constructor_action\x18\x03 \x01(\x0b\x32<.syft.core.node.common.action.RunFunctionOrConstructorActionH\x00\x12U\n\x17run_class_method_action\x18\x04 \x01(\x0b\x32\x32.syft.core.node.common.action.RunClassMethodActionH\x00\x12\x61\n\x1dgarbage_collect_object_action\x18\x06 \x01(\x0b\x32\x38.syft.core.node.common.action.GarbageCollectObjectActionH\x00\x12U\n\x15\x65num_attribute_action\x18\x07 \x01(\x0b\x32\x34.syft.core.node.common.action.GetEnumAttributeActionH\x00\x12Z\n\x1aget_or_set_property_action\x18\x08 \x01(\x0b\x32\x34.syft.core.node.common.action.GetOrSetPropertyActionH\x00\x12\x64\n\x1fget_set_static_attribute_action\x18\t \x01(\x0b\x32\x39.syft.core.node.common.action.GetSetStaticAttributeActionH\x00\x12L\n\x12save_object_action\x18\n \x01(\x0b\x32..syft.core.node.common.action.SaveObjectActionH\x00\x42\x08\n\x06\x61\x63tionb\x06proto3',
    dependencies=[
        proto_dot_core_dot_node_dot_common_dot_action_dot_get__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_run__function__or__constructor__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_run__class__method__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_garbage__collect__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_get__enum__attribute__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__property__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__static__attribute__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_action_dot_save__object__pb2.DESCRIPTOR,
    ],
)


_ACTION = _descriptor.Descriptor(
    name="Action",
    full_name="syft.core.node.common.action.Action",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="obj_type",
            full_name="syft.core.node.common.action.Action.obj_type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="get_object_action",
            full_name="syft.core.node.common.action.Action.get_object_action",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="run_function_or_constructor_action",
            full_name="syft.core.node.common.action.Action.run_function_or_constructor_action",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="run_class_method_action",
            full_name="syft.core.node.common.action.Action.run_class_method_action",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="garbage_collect_object_action",
            full_name="syft.core.node.common.action.Action.garbage_collect_object_action",
            index=4,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="enum_attribute_action",
            full_name="syft.core.node.common.action.Action.enum_attribute_action",
            index=5,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="get_or_set_property_action",
            full_name="syft.core.node.common.action.Action.get_or_set_property_action",
            index=6,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="get_set_static_attribute_action",
            full_name="syft.core.node.common.action.Action.get_set_static_attribute_action",
            index=7,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="save_object_action",
            full_name="syft.core.node.common.action.Action.save_object_action",
            index=8,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="action",
            full_name="syft.core.node.common.action.Action.action",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=525,
    serialized_end=1290,
)

_ACTION.fields_by_name[
    "get_object_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_get__object__pb2._GETOBJECTACTION
)
_ACTION.fields_by_name[
    "run_function_or_constructor_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_run__function__or__constructor__pb2._RUNFUNCTIONORCONSTRUCTORACTION
)
_ACTION.fields_by_name[
    "run_class_method_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_run__class__method__pb2._RUNCLASSMETHODACTION
)
_ACTION.fields_by_name[
    "garbage_collect_object_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_garbage__collect__object__pb2._GARBAGECOLLECTOBJECTACTION
)
_ACTION.fields_by_name[
    "enum_attribute_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_get__enum__attribute__pb2._GETENUMATTRIBUTEACTION
)
_ACTION.fields_by_name[
    "get_or_set_property_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__property__pb2._GETORSETPROPERTYACTION
)
_ACTION.fields_by_name[
    "get_set_static_attribute_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__static__attribute__pb2._GETSETSTATICATTRIBUTEACTION
)
_ACTION.fields_by_name[
    "save_object_action"
].message_type = (
    proto_dot_core_dot_node_dot_common_dot_action_dot_save__object__pb2._SAVEOBJECTACTION
)
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["get_object_action"]
)
_ACTION.fields_by_name["get_object_action"].containing_oneof = _ACTION.oneofs_by_name[
    "action"
]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["run_function_or_constructor_action"]
)
_ACTION.fields_by_name[
    "run_function_or_constructor_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["run_class_method_action"]
)
_ACTION.fields_by_name[
    "run_class_method_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["garbage_collect_object_action"]
)
_ACTION.fields_by_name[
    "garbage_collect_object_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["enum_attribute_action"]
)
_ACTION.fields_by_name[
    "enum_attribute_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["get_or_set_property_action"]
)
_ACTION.fields_by_name[
    "get_or_set_property_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["get_set_static_attribute_action"]
)
_ACTION.fields_by_name[
    "get_set_static_attribute_action"
].containing_oneof = _ACTION.oneofs_by_name["action"]
_ACTION.oneofs_by_name["action"].fields.append(
    _ACTION.fields_by_name["save_object_action"]
)
_ACTION.fields_by_name["save_object_action"].containing_oneof = _ACTION.oneofs_by_name[
    "action"
]
DESCRIPTOR.message_types_by_name["Action"] = _ACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Action = _reflection.GeneratedProtocolMessageType(
    "Action",
    (_message.Message,),
    {
        "DESCRIPTOR": _ACTION,
        "__module__": "proto.core.node.common.action.action_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.Action)
    },
)
_sym_db.RegisterMessage(Action)


# @@protoc_insertion_point(module_scope)



# syft absolute
import syft as sy


def test_union_pointer_creation() -> None:
    # check that we don't have this random type in our ast
    assert not hasattr(sy.lib_ast.syft.lib.misc.union, "BoolListUnion")

    union_pointer_type = sy.lib.misc.union.UnionGenerator[
        "syft.lib.python.Bool", "syft.lib.python.List"
    ]

    # for a union type to be generated in the ast, we need to rebuild it.
    new_ast = sy.lib.create_lib_ast()
    assert hasattr(new_ast.syft.lib.misc.union, "BoolListUnion")

    # the newly added type is visible in the global scope
    union_type = getattr(
        sy.lib.misc.union, union_pointer_type.rsplit(".", maxsplit=1)[-1]
    )

    # the union has all common allowed functions
    assert hasattr(union_type, "__add__")

    # the union respects List specific operations that don't interfere
    # with the Bool method set
    assert hasattr(union_type, "append")
    assert hasattr(union_type, "clear")

    # the union respects Bool specific operations that don't interfere
    # with the List method set
    assert hasattr(union_type, "__float__")
    assert hasattr(union_type, "__pow__")

    # inspecting that function generation is respected in the pointer type as
    # well, we care only about the pointer type. The union is acting as a
    # proxy to the original type, it's not going to be created.
    union_pointer_function_set = new_ast.syft.lib.misc.union.BoolListUnion.attrs.keys()
    bool_function_set = set(new_ast.syft.lib.python.Bool.attrs.keys())
    list_function_set = set(new_ast.syft.lib.python.List.attrs.keys())
    bool_list_function_set = bool_function_set.union(list_function_set)

    assert "__add__" in union_pointer_function_set
    assert "append" in union_pointer_function_set
    assert "clear" in union_pointer_function_set
    assert "__float__" in union_pointer_function_set
    assert "__pow__" in union_pointer_function_set

    # dummy example found out in the codebase of the behavior of a denied function
    # in the union type -> `__radd__`

    # function present in the Bool pointer
    assert "__radd__" in bool_list_function_set

    # function not present in the List pointer
    assert "__radd__" not in list_function_set

    # function present in the List function set
    assert "__radd__" in dir(sy.lib.python.List)

    # the __radd__ is denied from the List pointer function set
    # thus, neither the union pointer should have it due to possible anomalous
    # behaviors. If you want to use it on Bool, you will need to cast it.
    assert "__radd__" not in union_pointer_function_set



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import pydp
from pydp.algorithms.laplacian import BoundedMean
from pydp.algorithms.laplacian import BoundedStandardDeviation
from pydp.algorithms.laplacian import BoundedSum
from pydp.algorithms.laplacian import BoundedVariance
from pydp.algorithms.laplacian import Count
from pydp.algorithms.laplacian import Max
from pydp.algorithms.laplacian import Median
from pydp.algorithms.laplacian import Min
from pydp.algorithms.laplacian import Percentile

# syft relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..misc.union import UnionGenerator
from ..util import generic_update_ast

LIB_NAME = "pydp"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "python": {"max_version": (3, 8, 99)},
}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("pydp", pydp),
        ("pydp.algorithms", pydp.algorithms),
        ("pydp.algorithms.laplacian", pydp.algorithms.laplacian),
    ]
    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        (
            "pydp.algorithms.laplacian.BoundedMean",
            "pydp.algorithms.laplacian.BoundedMean",
            BoundedMean,
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum",
            "pydp.algorithms.laplacian.BoundedSum",
            BoundedSum,
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation",
            "pydp.algorithms.laplacian.BoundedStandardDeviation",
            BoundedStandardDeviation,
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance",
            "pydp.algorithms.laplacian.BoundedVariance",
            BoundedVariance,
        ),
        (
            "pydp.algorithms.laplacian.Min",
            "pydp.algorithms.laplacian.Min",
            Min,
        ),
        (
            "pydp.algorithms.laplacian.Max",
            "pydp.algorithms.laplacian.Max",
            Max,
        ),
        (
            "pydp.algorithms.laplacian.Median",
            "pydp.algorithms.laplacian.Median",
            Median,
        ),
        (
            "pydp.algorithms.laplacian.Percentile",
            "pydp.algorithms.laplacian.Percentile",
            Percentile,
        ),
        (
            "pydp.algorithms.laplacian.Count",
            "pydp.algorithms.laplacian.Count",
            Count,
        ),
    ]

    methods = [
        (
            "pydp.algorithms.laplacian.BoundedMean.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.BoundedMean.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.BoundedSum.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.BoundedStandardDeviation.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.BoundedVariance.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Min.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Min.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Min.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Min.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Min.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Min.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.Min.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Max.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Max.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Max.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Max.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Max.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Max.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.Max.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Median.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Median.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Median.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Median.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Median.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Median.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.Median.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Percentile.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.Percentile.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Count.quick_result",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Count.add_entries",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Count.add_entry",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Count.privacy_budget_left",
            "syft.lib.python.Float",
        ),
        (
            "pydp.algorithms.laplacian.Count.reset",
            "syft.lib.python._SyNone",
        ),
        (
            "pydp.algorithms.laplacian.Count.result",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"],
        ),
        (
            "pydp.algorithms.laplacian.Count.noise_confidence_interval",
            "syft.lib.python.Float",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



"""
This is the main Plan class which is responsible for containing a list of Actions
which can be serialized, deserialized, and executed by substituting the run time
pointers with the original traced pointers and replaying the actions against a node.
"""
# stdlib
import re
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ... import serialize
from ...logger import traceback_and_raise
from ...proto.core.node.common.action.action_pb2 import Action as Action_PB
from ...proto.core.plan.plan_pb2 import Plan as Plan_PB
from ..common.object import Serializable
from ..common.serde.serializable import bind_protobuf
from ..node.abstract.node import AbstractNode
from ..node.common import client
from ..node.common.action.common import Action
from ..node.common.util import listify
from ..pointer.pointer import Pointer
from ..store.storeable_object import StorableObject

CAMEL_TO_SNAKE_PAT = re.compile(r"(?<!^)(?=[A-Z])")


@bind_protobuf
class Plan(Serializable):
    """
    A plan is a collection of actions, plus some variable inputs, that together form a computation graph.

    Attributes:
        actions: list of actions
        inputs: Pointers to the inputs. Defaults to None.
    """

    def __init__(
        self,
        actions: Union[List[Action], None] = None,
        inputs: Union[Dict[str, Pointer], None] = None,
        outputs: Union[Pointer, List[Pointer], None] = None,
        code: Optional[str] = None,
        max_calls: Optional[int] = None,
    ):
        """
        Initialize the Plan with actions, inputs and outputs
        """
        self.actions: List[Action] = listify(actions)
        self.inputs: Dict[str, Pointer] = inputs if inputs is not None else dict()
        self.outputs: List[Pointer] = listify(outputs)
        self.code = code
        self.max_calls = max_calls
        self.n_calls = 0

    def __call__(
        self,
        node: Optional[AbstractNode] = None,
        verify_key: VerifyKey = None,
        **kwargs: Dict[str, Any],
    ) -> List[StorableObject]:
        """
        1) For all pointers that were passed into the init as `inputs`, this method
           replaces those pointers in self.actions by the pointers passed in as *args.
        2) Executes the actions in self.actions one by one

        *While this function requires `node` and `verify_key` as inputs, during remote
        execution, passing these is handled in `RunClassMethodAction`*

        *Note that this method will receive *args as pointers during execution. Normally,
        pointers are resolved during `RunClassMethodAction.execute()`, but not for plans,
        as they need to operate on the pointer to enable remapping of the inputs.*
        Args:
            *args: the new inputs for the plan, passed as pointers
        """

        self.n_calls += 1

        # this is pretty cumbersome, we are searching through all actions to check
        # if we need to redefine some of their attributes that are inputs in the
        # graph of actions
        if node is None:
            return self.execute_locally(**kwargs)

        new_inputs: Dict[str, Pointer] = {}
        for k, current_input in self.inputs.items():
            new_input = kwargs[k]
            if not issubclass(type(new_input), Pointer):
                traceback_and_raise(
                    f"Calling Plan without a Pointer. {k} == {type(new_input)} "
                )
            for a in self.actions:
                if hasattr(a, "remap_input"):
                    a.remap_input(current_input, new_input)  # type: ignore

            # redefine the inputs of the plan
            new_inputs[k] = new_input  # type: ignore
        self.inputs = new_inputs

        for a in self.actions:
            a.execute_action(node, verify_key)

        if len(self.outputs):
            resolved_outputs = []
            for arg in self.outputs:
                r_arg = node.store[arg.id_at_location]
                resolved_outputs.append(r_arg.data)
            return resolved_outputs
        else:
            return []

    def __repr__(self) -> str:
        obj_str = "Plan"

        allowed, remaining = (
            (self.max_calls, self.max_calls - self.n_calls)
            if self.max_calls is not None
            else ("not defined", "not defined")
        )

        ex_str = f"Allowed executions:\t{allowed}\nRemaining executions:\t{remaining}"

        inp_str = "Inputs:\n"
        inp_str += "\n".join(
            [f"\t\t{k}:\t{v.__class__.__name__}" for k, v in self.inputs.items()]
        )

        act_str = f"Actions:\n\t\t{len(self.actions)} Actions"

        out_str = "Outputs:\n"
        out_str += "\n".join([f"\t\t{o.__class__.__name__}" for o in self.outputs])

        plan_str = "Plan code:\n"
        plan_str += f'"""\n{self.code}\n"""' if self.code is not None else ""

        return f"{obj_str}\n{ex_str}\n{inp_str}\n{act_str}\n{out_str}\n\n{plan_str}"

    def execute_locally(self, **kwargs: Any) -> List[StorableObject]:
        """Execute a plan by sending it to a virtual machine and calling execute on the pointer.
        This is a workaround until we have a way to execute plans locally.
        """
        # prevent circular dependency
        # syft relative
        from ...core.node.vm.vm import VirtualMachine  # noqa: F401

        alice = VirtualMachine(name="plan_executor")
        alice_client: client.Client = alice.get_client()
        self_ptr = self.send(alice_client)  # type: ignore
        out = self_ptr(**kwargs)
        return out.get()

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return Plan_PB

    def _object2proto(self) -> Plan_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectWithID_PB

        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        def camel_to_snake(s: str) -> str:
            """Convert CamelCase classes to snake case for matching protobuf names"""
            return CAMEL_TO_SNAKE_PAT.sub("_", s).lower()

        actions_pb = [
            Action_PB(
                obj_type=".".join([action.__module__, action.__class__.__name__]),
                **{camel_to_snake(action.__class__.__name__): serialize(action)},
            )
            for action in self.actions
        ]
        inputs_pb = {k: v._object2proto() for k, v in self.inputs.items()}
        outputs_pb = [out._object2proto() for out in self.outputs]

        return Plan_PB(actions=actions_pb, inputs=inputs_pb, outputs=outputs_pb)

    @staticmethod
    def _proto2object(proto: Plan_PB) -> "Plan":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of Plan
        :rtype: Plan

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        actions = []

        for action_proto in proto.actions:
            module, cls_name = action_proto.obj_type.rsplit(".", 1)
            action_cls = getattr(sys.modules[module], cls_name)

            # protobuf does no inheritance, so we wrap action subclasses
            # in the main action class.
            inner_action = getattr(action_proto, action_proto.WhichOneof("action"))
            actions.append(action_cls._proto2object(inner_action))

        inputs = {k: Pointer._proto2object(proto.inputs[k]) for k in proto.inputs}
        outputs = [
            Pointer._proto2object(pointer_proto) for pointer_proto in proto.outputs
        ]

        return Plan(actions=actions, inputs=inputs, outputs=outputs)



# stdlib

# third party
import torch as th
from torch.nn import Parameter

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.torch.tensor_util import protobuf_tensor_deserializer
from ...lib.torch.tensor_util import protobuf_tensor_serializer
from ...proto.lib.torch.parameter_pb2 import ParameterProto as Parameter_PB

torch_tensor = th.tensor([1.0, 2.0, 3.0])
torch_parameter_type = type(th.nn.parameter.Parameter(torch_tensor))


def object2proto(obj: object) -> Parameter_PB:
    proto = Parameter_PB()
    tensor_data = getattr(obj, "data", None)
    if tensor_data is not None:
        proto.tensor.CopyFrom(protobuf_tensor_serializer(tensor_data))
    proto.requires_grad = getattr(obj, "requires_grad", False)
    grad = getattr(obj, "grad", None)
    if grad is not None:
        proto.grad.CopyFrom(protobuf_tensor_serializer(grad))

    # opacus monkey patches this onto the Parameter class
    grad_sample = getattr(obj, "grad_sample", None)
    if grad_sample is not None:
        proto.grad_sample.CopyFrom(protobuf_tensor_serializer(grad_sample))
    return proto


def proto2object(proto: Parameter_PB) -> Parameter:
    data = protobuf_tensor_deserializer(proto.tensor)
    param = Parameter(data, requires_grad=proto.requires_grad)
    if proto.HasField("grad"):
        param.grad = protobuf_tensor_deserializer(proto.grad)

    # opacus monkey patches this onto the Parameter class
    if proto.HasField("grad_sample"):
        param.grad_sample = protobuf_tensor_deserializer(proto.grad_sample)
    return param


GenerateWrapper(
    wrapped_type=torch_parameter_type,
    import_path="torch.nn.parameter.Parameter",
    protobuf_scheme=Parameter_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
from functools import partial
from typing import Any
from typing import List
from typing import Union

# third party
import numpy as np
import torch

# syft absolute
from syft.lib.tensor.tensorbase_util import call_func_and_wrap_result

Num = Union[int, float]


class ChildDelegatorTensor:
    def __getattr__(self, name: str) -> Any:
        func_or_attr = getattr(self.child, name)
        if callable(func_or_attr):
            return partial(call_func_and_wrap_result, func_or_attr, self.__class__)
        else:
            var = func_or_attr
            return var


class DataTensor(ChildDelegatorTensor):
    def __init__(self, child: Union[torch.FloatTensor, torch.IntTensor]) -> None:
        self.child = child

    def __add__(self, other: "DataTensor") -> "DataTensor":
        return DataTensor(child=self.child + other.child)

    def __sub__(self, other: "DataTensor") -> "DataTensor":
        return DataTensor(child=self.child - other.child)

    def __mul__(self, other: "DataTensor") -> "DataTensor":
        return DataTensor(child=self.child * other.child)

    def __truediv__(self, other: Union[int, float]) -> "DataTensor":
        return DataTensor(child=self.child / other)

    def __matmul__(self, other: "DataTensor") -> "DataTensor":
        return DataTensor(child=self.child @ other.child)


class FloatTensor(ChildDelegatorTensor):
    def __init__(self, child: DataTensor) -> None:
        self.child = child

    def __add__(self, other: "FloatTensor") -> "FloatTensor":
        return FloatTensor(child=self.child + other.child)

    def __sub__(self, other: "FloatTensor") -> "FloatTensor":
        return FloatTensor(child=self.child - other.child)

    def __mul__(self, other: "FloatTensor") -> "FloatTensor":
        return FloatTensor(child=self.child * other.child)

    def __truediv__(self, other: Union[int, float]) -> "FloatTensor":
        return FloatTensor(child=self.child / other)

    def __matmul__(self, other: "FloatTensor") -> "FloatTensor":
        return FloatTensor(child=self.child @ other.child)


class IntegerTensor(ChildDelegatorTensor):
    def __init__(self, child: DataTensor) -> None:
        self.child = child

    # todo: make sure that operations return the correct tensor type
    # e.g. a div between a IntegerTensor and an Int may produce
    # a FloatTensor

    def __add__(self, other: "IntegerTensor") -> "IntegerTensor":
        return IntegerTensor(child=self.child + other.child)

    def __sub__(self, other: "IntegerTensor") -> "IntegerTensor":
        return IntegerTensor(child=self.child - other.child)

    def __mul__(self, other: "IntegerTensor") -> "IntegerTensor":
        return IntegerTensor(child=self.child * other.child)

    def __truediv__(self, other: Union[int, float]) -> "IntegerTensor":
        return IntegerTensor(child=self.child / other)

    def __matmul__(self, other: "IntegerTensor") -> "IntegerTensor":
        return IntegerTensor(child=self.child @ other.child)


class SyftTensor(ChildDelegatorTensor):
    def __init__(self, child: Union[FloatTensor, IntegerTensor]) -> None:
        self.child = child

    def __add__(self, other: "SyftTensor") -> "SyftTensor":
        if isinstance(self.child, FloatTensor) and isinstance(other.child, FloatTensor):
            return SyftTensor(child=self.child + other.child)
        elif isinstance(self.child, IntegerTensor) and isinstance(
            other.child, IntegerTensor
        ):
            return SyftTensor(child=self.child + other.child)
        else:
            raise ValueError()

    def __sub__(self, other: "SyftTensor") -> "SyftTensor":
        if isinstance(self.child, FloatTensor) and isinstance(other.child, FloatTensor):
            return SyftTensor(child=self.child - other.child)
        elif isinstance(self.child, IntegerTensor) and isinstance(
            other.child, IntegerTensor
        ):
            return SyftTensor(child=self.child - other.child)
        else:
            raise ValueError()

    def __mul__(self, other: "SyftTensor") -> "SyftTensor":
        if isinstance(self.child, FloatTensor) and isinstance(other.child, FloatTensor):
            return SyftTensor(child=self.child * other.child)
        elif isinstance(self.child, IntegerTensor) and isinstance(
            other.child, IntegerTensor
        ):
            return SyftTensor(child=self.child * other.child)
        else:
            raise ValueError()

    def __truediv__(self, other: Union[int, float]) -> "SyftTensor":
        if isinstance(self.child, FloatTensor) and isinstance(other, int):
            return SyftTensor(child=self.child / other)
        if isinstance(self.child, IntegerTensor) and isinstance(other, float):
            return SyftTensor(child=self.child / other)
        else:
            raise ValueError()

    def __matmul__(self, other: "SyftTensor") -> "SyftTensor":
        if (
            isinstance(self.child, FloatTensor)
            and isinstance(other, SyftTensor)
            and isinstance(other.child, FloatTensor)
        ):
            return SyftTensor(child=self.child @ other.child)
        if (
            isinstance(self.child, IntegerTensor)
            and isinstance(other, SyftTensor)
            and isinstance(other.child, IntegerTensor)
        ):
            return SyftTensor(child=self.child @ other.child)
        else:
            raise ValueError()

    @classmethod
    def FloatTensor(cls, data: Union[List[Num], np.ndarray]) -> "SyftTensor":
        if isinstance(data, list) or isinstance(data, np.ndarray):
            return cls(
                child=FloatTensor(child=DataTensor(child=torch.FloatTensor(data)))
            )
        else:
            raise NotImplementedError()

    @classmethod
    def IntegerTensor(cls, data: Union[List[Num], np.ndarray]) -> "SyftTensor":
        if isinstance(data, list) or isinstance(data, np.ndarray):
            return cls(
                child=IntegerTensor(child=DataTensor(child=torch.IntTensor(data)))
            )
        else:
            raise NotImplementedError()



# coding: utf-8
# stdlib
import argparse
import math
import os
import time

# third party
import torch
import torch.nn as nn
import torch.onnx

import data  # isort:skip
import model  # isort:skip

parser = argparse.ArgumentParser(
    description="PyTorch Wikitext-2 RNN/LSTM/GRU/Transformer Language Model"
)
parser.add_argument(
    "--data",
    type=str,
    default="./data/wikitext-2",
    help="location of the data corpus; default: \"./data/wikitext-2\""
)
parser.add_argument(
    "--model",
    type=str,
    default="LSTM",
    help="type of recurrent net (RNN_TANH, RNN_RELU, LSTM, GRU, Transformer); default: \"LSTM\"",
)
parser.add_argument(
    "--emsize",
    type=int,
    default=200,
    help="size of word embeddings; default: 200"
)
parser.add_argument(
    "--nhid",
    type=int,
    default=200,
    help="number of hidden units per layer; default: 200"
)
parser.add_argument(
    "--nlayers",
    type=int,
    default=2,
    help="number of layers; default: 2"
)
parser.add_argument(
    "--lr",
    type=float,
    default=20,
    help="initial learning rate; default: 20"
)
parser.add_argument(
    "--clip",
    type=float,
    default=0.25,
    help="gradient clipping; default: 0.25"
)
parser.add_argument(
    "--epochs",
    type=int,
    default=40,
    help="upper epoch limit; default: 40"
)
parser.add_argument(
    "--batch_size",
    type=int,
    default=20,
    metavar="N",
    help="batch size; default: 20"
)
parser.add_argument(
    "--bptt",
    type=int,
    default=35,
    help="sequence length; default: 35"
)
parser.add_argument(
    "--dropout",
    type=float,
    default=0.2,
    help="dropout applied to layers (0 = no dropout); default: 0.2",
)
parser.add_argument(
    "--tied",
    action="store_true",
    help="tie the word embedding and softmax weights"
)
parser.add_argument(
    "--seed",
    type=int,
    default=1111,
    help="random seed; default: 1111"
)
parser.add_argument(
    "--cuda",
    action="store_true",
    help="use CUDA"
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=200,
    metavar="N",
    help="report interval; default: 200"
)
parser.add_argument(
    "--save",
    type=str,
    default="model.pt",
    help="path to save the final model; default: \"model.pt\""
)
parser.add_argument(
    "--onnx-export",
    type=str,
    default="",
    help="path to export the final model in onnx format; default: ''",
)

parser.add_argument(
    "--nhead",
    type=int,
    default=2,
    help="the number of heads in the encoder/decoder of the transformer model; default: 2",
)
parser.add_argument(
    "--dry-run",
    action="store_true",
    help="verify the code and the model"
)

args = parser.parse_args()

# Set the random seed manually for reproducibility.
torch.manual_seed(args.seed)
if torch.cuda.is_available():
    if not args.cuda:
        print("WARNING: You have a CUDA device, so you should probably run with --cuda")

device = torch.device("cuda" if args.cuda else "cpu")


###############################################################################
# Load data
###############################################################################

corpus = data.Corpus(args.data)

# Starting from sequential data, batchify arranges the dataset into columns.
# For instance, with the alphabet as the sequence and batch size 4, we'd get
# ┌ a g m s ┐
# │ b h n t │
# │ c i o u │
# │ d j p v │
# │ e k q w │
# └ f l r x ┘.
# These columns are treated as independent by the model, which means that the
# dependence of e. g. 'g' on 'f' can not be learned, but allows more efficient
# batch processing.


def batchify(data, bsz):
    # Work out how cleanly we can divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    return data.to(device)


eval_batch_size = 10
train_data = batchify(corpus.train, args.batch_size)
val_data = batchify(corpus.valid, eval_batch_size)
test_data = batchify(corpus.test, eval_batch_size)


###############################################################################
# Build the model
###############################################################################

ntokens = len(corpus.dictionary)
if args.model == "Transformer":
    model = model.TransformerModel(
        ntokens, args.emsize, args.nhead, args.nhid, args.nlayers, args.dropout
    ).to(device)
else:
    model = model.RNNModel(
        args.model,
        ntokens,
        args.emsize,
        args.nhid,
        args.nlayers,
        args.dropout,
        args.tied,
    ).to(device)

criterion = nn.NLLLoss()


###############################################################################
# Training code
###############################################################################

def repackage_hidden(h):
    """Wraps hidden states in new Tensors, to detach them from their history."""

    if isinstance(h, torch.Tensor):
        return h.detach()
    else:
        return tuple(repackage_hidden(v) for v in h)


# get_batch subdivides the source data into chunks of length args.bptt.
# If source is equal to the example output of the batchify function, with
# a bptt-limit of 2, we'd get the following two Variables for i = 0:
# ┌ a g m s ┐ ┌ b h n t ┐
# └ b h n t ┘ └ c i o u ┘
# Note that despite the name of the function, the subdivison of data is not
# done along the batch dimension (i.e. dimension 1), since that was handled
# by the batchify function. The chunks are along dimension 0, corresponding
# to the seq_len dimension in the LSTM.


def get_batch(source, i):
    seq_len = min(args.bptt, len(source) - 1 - i)
    data = source[i : i + seq_len]
    target = source[i + 1 : i + 1 + seq_len].view(-1)
    return data, target


def evaluate(data_source):
    # Turn on evaluation mode which disables dropout.
    model.eval()
    total_loss = 0.0
    ntokens = len(corpus.dictionary)
    if args.model != "Transformer":
        hidden = model.init_hidden(eval_batch_size)
    with torch.no_grad():
        for i in range(0, data_source.size(0) - 1, args.bptt):
            data, targets = get_batch(data_source, i)
            if args.model == "Transformer":
                output = model(data)
                output = output.view(-1, ntokens)
            else:
                output, hidden = model(data, hidden)
                hidden = repackage_hidden(hidden)
            total_loss += len(data) * criterion(output, targets).item()
    return total_loss / (len(data_source) - 1)


def train():
    # Turn on training mode which enables dropout.
    model.train()
    total_loss = 0.0
    start_time = time.time()
    ntokens = len(corpus.dictionary)
    if args.model != "Transformer":
        hidden = model.init_hidden(args.batch_size)
    for batch, i in enumerate(range(0, train_data.size(0) - 1, args.bptt)):
        data, targets = get_batch(train_data, i)
        # Starting each batch, we detach the hidden state from how it was previously produced.
        # If we didn't, the model would try backpropagating all the way to start of the dataset.
        model.zero_grad()
        if args.model == "Transformer":
            output = model(data)
            output = output.view(-1, ntokens)
        else:
            hidden = repackage_hidden(hidden)
            output, hidden = model(data, hidden)
        loss = criterion(output, targets)
        loss.backward()

        # `clip_grad_norm` helps prevent the exploding gradient problem in RNNs / LSTMs.
        torch.nn.utils.clip_grad_norm_(model.parameters(), args.clip)
        for p in model.parameters():
            p.data.add_(p.grad, alpha=-lr)

        total_loss += loss.item()

        if batch % args.log_interval == 0 and batch > 0:
            cur_loss = total_loss / args.log_interval
            elapsed = time.time() - start_time
            print(
                "| epoch {:3d} | {:5d}/{:5d} batches | lr {:02.2f} | ms/batch {:5.2f} | "
                "loss {:5.2f} | ppl {:8.2f}".format(
                    epoch,
                    batch,
                    len(train_data) // args.bptt,
                    lr,
                    elapsed * 1000 / args.log_interval,
                    cur_loss,
                    math.exp(cur_loss),
                )
            )
            total_loss = 0
            start_time = time.time()
        if args.dry_run:
            break


def export_onnx(path, batch_size, seq_len):
    print(
        "The model is also exported in ONNX format at {}".format(
            os.path.realpath(args.onnx_export)
        )
    )
    model.eval()
    dummy_input = (
        torch.LongTensor(seq_len * batch_size).zero_().view(-1, batch_size).to(device)
    )
    hidden = model.init_hidden(batch_size)
    torch.onnx.export(model, (dummy_input, hidden), path)


# Loop over epochs.
lr = args.lr
best_val_loss = None

# At any point you can hit Ctrl + C to break out of training early.
try:
    for epoch in range(1, args.epochs + 1):
        epoch_start_time = time.time()
        train()
        val_loss = evaluate(val_data)
        print("-" * 89)
        print(
            "| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | "
            "valid ppl {:8.2f}".format(
                epoch, (time.time() - epoch_start_time), val_loss, math.exp(val_loss)
            )
        )
        print("-" * 89)
        # Save the model if the validation loss is the best we've seen so far.
        if not best_val_loss or val_loss < best_val_loss:
            with open(args.save, "wb") as f:
                torch.save(model, f)
            best_val_loss = val_loss
        else:
            # Anneal the learning rate if no improvement has been seen in the validation dataset.
            lr /= 4.0
except KeyboardInterrupt:
    print("-" * 89)
    print("Exiting from training early")

# Load the best saved model.
with open(args.save, "rb") as f:
    model = torch.load(f)
    # after load the rnn params are not a continuous chunk of memory
    # this makes them a continuous chunk, and will speed up forward pass
    # Currently, only rnn model supports flatten_parameters function.
    if args.model in ["RNN_TANH", "RNN_RELU", "LSTM", "GRU"]:
        model.rnn.flatten_parameters()

# Run on test data.
test_loss = evaluate(test_data)
print("=" * 89)
print(
    "| End of training | test loss {:5.2f} | test ppl {:8.2f}".format(
        test_loss, math.exp(test_loss)
    )
)
print("=" * 89)

if len(args.onnx_export) > 0:
    # Export the model in ONNX format.
    export_onnx(args.onnx_export, batch_size=1, seq_len=args.bptt)



# syft absolute
import syft as sy
from syft import serialize
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.node.common.action.get_object_action import GetObjectAction


def test_get_object_action_serde() -> None:
    msg = GetObjectAction(
        id_at_location=UID(), address=Address(), reply_to=Address(), msg_id=UID()
    )
    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.id_at_location == msg2.id_at_location
    assert msg.address == msg2.address
    assert msg.reply_to == msg2.reply_to



# stdlib
from typing import Any
from typing import Type

# third party
from google.protobuf.empty_pb2 import Empty as Empty_PB
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.core.auth.signed_message_pb2 import VerifyAll as VerifyAllWrapper_PB
from ...proto.core.auth.signed_message_pb2 import VerifyKey as VerifyKey_PB
from .serde.serializable import Serializable
from .serde.serializable import bind_protobuf


def object2proto(obj: Any) -> VerifyKey_PB:
    return VerifyKey_PB(verify_key=bytes(obj))


def proto2object(proto: VerifyKey_PB) -> VerifyKey:
    return VerifyKey(proto.verify_key)


GenerateWrapper(
    wrapped_type=VerifyKey,
    import_path="nacl.signing.VerifyKey",
    protobuf_scheme=VerifyKey_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)


def _create_VERIFYALL() -> Any:
    @bind_protobuf
    class VerifyAll(Serializable):
        _instance = None

        def __new__(cls: Type) -> "VerifyAll":
            if cls._instance is None:
                cls._instance = object.__new__(cls)
            return cls._instance

        def _object2proto(self) -> VerifyAllWrapper_PB:
            return VerifyAllWrapper_PB(all=Empty_PB())

        @staticmethod
        def _proto2object(proto: VerifyAllWrapper_PB) -> "VerifyAll":
            return VERIFYALL

        @staticmethod
        def get_protobuf_schema() -> GeneratedProtocolMessageType:
            return VerifyAllWrapper_PB

    return VerifyAll()


VERIFYALL = _create_VERIFYALL()
VerifyAll = type(
    VERIFYALL
)  # deprecated: https://github.com/OpenMined/PySyft/issues/5396



# stdlib
from collections import UserDict

# third party
import pytest
import torch as th

# syft absolute
import syft as sy
from syft.core.common.uid import UID
from syft.lib.python.dict import Dict
from syft.lib.python.int import Int
from syft.lib.python.string import String
from syft.proto.lib.python.dict_pb2 import Dict as Dict_PB


def test_dict_creation() -> None:
    d1 = {String("t1"): 1, String("t2"): 2}
    dict1 = Dict(d1)
    assert type(getattr(dict1, "id", None)) is UID

    d2 = dict({"t1": 1, "t2": 2})
    dict2 = Dict(d2)
    dict2._id = UID()
    assert type(getattr(dict2, "id", None)) is UID

    d3 = UserDict({"t1": 1, "t2": 2})
    dict3 = Dict(**d3)
    assert type(getattr(dict3, "id", None)) is UID

    assert dict1.keys() == dict2.keys()
    assert dict1.keys() == dict3.keys()

    # ValuesView uses object.__eq__
    # https://stackoverflow.com/questions/34312674/why-are-the-values-of-an-ordereddict-not-equal
    assert dict1.values() != dict2.values()
    assert dict1.values() != dict3.values()

    assert dict1.items() == dict2.items()
    assert dict1.items() == dict3.items()

    it = list(iter(dict2.values()))
    assert len(it) == 2
    assert type(it) is list


def test_dict_serde() -> None:
    t1 = th.tensor([1, 2])
    t2 = th.tensor([1, 3])

    syft_list = Dict({Int(1): t1, Int(2): t2})
    assert type(getattr(syft_list, "id", None)) is UID

    serialized = syft_list._object2proto()

    assert isinstance(serialized, Dict_PB)

    deserialized = Dict._proto2object(proto=serialized)

    assert isinstance(deserialized, Dict)
    assert deserialized.id == syft_list.id
    for deserialized_el, original_el in zip(deserialized, syft_list):
        assert deserialized_el == original_el


def test_list_send(client: sy.VirtualMachineClient) -> None:
    syft_list = Dict({String("t1"): String("test"), String("t2"): String("test")})
    ptr = syft_list.send(client)
    # Check pointer type
    assert ptr.__class__.__name__ == "DictPointer"

    # Check that we can get back the object
    res = ptr.get()
    for res_el, original_el in zip(res, syft_list):
        assert res_el == original_el


@pytest.mark.parametrize("method_name", ["items", "keys", "values"])
def test_iterator_methods(method_name: str, root_client) -> None:
    d = Dict({"#1": 1, "#2": 2})
    dptr = d.send(root_client)

    itemsptr = getattr(dptr, method_name)()
    assert type(itemsptr).__name__ == "IteratorPointer"

    for itemptr, local_item in zip(itemsptr, getattr(d, method_name)()):
        get_item = itemptr.get()
        assert get_item == local_item



# stdlib
import argparse
from collections import namedtuple
from itertools import count

# third party
import gym
import numpy as np
import torch
from torch.distributions import Categorical
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# Cart Pole
parser = argparse.ArgumentParser(description="PyTorch actor-critic example")
parser.add_argument(
    "--gamma",
    type=float,
    default=0.99,
    metavar="G",
    help="discount factor (default: 0.99)",
)
parser.add_argument(
    "--seed",
    type=int,
    default=543,
    metavar="N",
    help="random seed (default: 543)"
)
parser.add_argument(
    "--render",
    action="store_true",
    help="render the environment"
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=10,
    metavar="N",
    help="interval between training status logs (default: 10)",
)
args = parser.parse_args()


env = gym.make("CartPole-v0")
env.seed(args.seed)
torch.manual_seed(args.seed)


SavedAction = namedtuple("SavedAction", ["log_prob", "value"])


class Policy(nn.Module):
    """
    implements both actor and critic in one model
    """

    def __init__(self):
        super(Policy, self).__init__()
        self.affine1 = nn.Linear(4, 128)

        # actor's layer
        self.action_head = nn.Linear(128, 2)

        # critic's layer
        self.value_head = nn.Linear(128, 1)

        # action & reward buffer
        self.saved_actions = []
        self.rewards = []

    def forward(self, x):
        """
        forward of both actor and critic
        """
        x = F.relu(self.affine1(x))

        # actor: choses action to take from state s_t
        # by returning probability of each action
        action_prob = F.softmax(self.action_head(x), dim=-1)

        # critic: evaluates being in the state s_t
        state_values = self.value_head(x)

        # return values for both actor and critic as a tuple of 2 values:
        # 1. a list with the probability of each action over the action space
        # 2. the value from state s_t
        return action_prob, state_values


model = Policy()
optimizer = optim.Adam(model.parameters(), lr=3e-2)
eps = np.finfo(np.float32).eps.item()


def select_action(state):
    state = torch.from_numpy(state).float()
    probs, state_value = model(state)

    # create a categorical distribution over the list of probabilities of actions
    m = Categorical(probs)

    # and sample an action using the distribution
    action = m.sample()

    # save to action buffer
    model.saved_actions.append(SavedAction(m.log_prob(action), state_value))

    # the action to take (left or right)
    return action.item()


def finish_episode():
    """
    Training code. Calculates actor and critic loss and performs backpropagation.
    """
    R = 0
    saved_actions = model.saved_actions
    policy_losses = []  # list to save actor (policy) loss
    value_losses = []  # list to save critic (value) loss
    returns = []  # list to save the true values

    # calculate the true value using rewards returned from the environment
    for r in model.rewards[::-1]:
        # calculate the discounted value
        R = r + args.gamma * R
        returns.insert(0, R)

    returns = torch.tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + eps)

    for (log_prob, value), R in zip(saved_actions, returns):
        advantage = R - value.item()

        # calculate actor (policy) loss
        policy_losses.append(-log_prob * advantage)

        # calculate critic (value) loss using L1 smooth loss
        value_losses.append(F.smooth_l1_loss(value, torch.tensor([R])))

    # reset gradients
    optimizer.zero_grad()

    # sum up all the values of policy_losses and value_losses
    loss = torch.stack(policy_losses).sum() + torch.stack(value_losses).sum()

    # perform backprop
    loss.backward()
    optimizer.step()

    # reset rewards and action buffer
    del model.rewards[:]
    del model.saved_actions[:]


def main():
    running_reward = 10

    # run inifinitely many episodes
    for i_episode in count(1):

        # reset environment and episode reward
        state = env.reset()
        ep_reward = 0

        # for each episode, only run 9999 steps so that we don't
        # infinite loop while learning
        for t in range(1, 10000):

            # select action from policy
            action = select_action(state)

            # take the action
            state, reward, done, _ = env.step(action)

            if args.render:
                env.render()

            model.rewards.append(reward)
            ep_reward += reward
            if done:
                break

        # update cumulative reward
        running_reward = 0.05 * ep_reward + (1 - 0.05) * running_reward

        # perform backprop
        finish_episode()

        # log results
        if i_episode % args.log_interval == 0:
            print(
                "Episode {}\tLast reward: {:.2f}\tAverage reward: {:.2f}".format(
                    i_episode, ep_reward, running_reward
                )
            )

        # check if we have "solved" the cart pole problem
        if running_reward > env.spec.reward_threshold:
            print(
                "Solved! Running reward is now {} and "
                "the last episode runs to {} time steps!".format(running_reward, t)
            )
            break


if __name__ == "__main__":
    main()



# stdlib
from typing import List

# syft absolute
from syft.lib.python import List as SyList
from syft.lib.python.string import String


def string_serde(data: str) -> None:
    syft_string = String(data)

    serialized = syft_string._object2proto()
    String._proto2object(proto=serialized)


def list_serde(data: List[str]) -> None:
    syft_list = SyList(data)

    serialized = syft_list._object2proto()
    SyList._proto2object(proto=serialized)



# stdlib
import io
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from syft_proto.execution.v1.plan_pb2 import Plan as Plan_PB
import torch as th

# syft relative
from .....core.common.object import Serializable
from .....logger import traceback_and_raise


class PlanTorchscript(Serializable):
    """
    Represents Syft Plan translated to TorchScript
    """

    def __init__(self, torchscript: Optional[th.jit.ScriptModule] = None):
        super().__init__()
        self.torchscript = torchscript

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """
        Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        Returns:
            The type of protobuf object which corresponds to this class.

        """
        return Plan_PB

    def __call__(self, *args: Any) -> Any:
        if self.torchscript:
            return self.torchscript(*args)
        else:
            traceback_and_raise("No torchscript")

    def _object2proto(self) -> Plan_PB:
        """
        Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        Returns:
            The protobuf representation of self.

        """
        bin = self.torchscript.save_to_buffer() if self.torchscript else None
        return Plan_PB(torchscript=bin)

    @staticmethod
    def _proto2object(proto: Plan_PB) -> "PlanTorchscript":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        Returns:
           A Plan representation from the associated protobuf message.
        """
        if proto.torchscript:
            bin = io.BytesIO(proto.torchscript)
            ts = th.jit.load(bin)
        else:
            ts = None

        return PlanTorchscript(torchscript=ts)



# run this command line like:
# python examples/duet/cli/do_no_threads.py

# WARNING if this is run in REPL / Interactive it will block so you need to figure
# out a solution to that separately

# stdlib
import asyncio
import os

os.environ["SYFT_USE_EVENT_LOOP_THREAD"] = "0"
loop = asyncio.new_event_loop()
asyncio._set_running_loop(loop)

# syft absolute
import syft as sy

duet = sy.launch_duet(loopback=True)

print("DO Store", duet.store)



# syft absolute
import syft as sy
from syft.lib.python.float import Float
from syft.proto.lib.python.float_pb2 import Float as Float_PB


def test_serde() -> None:
    syft_float = Float(5)

    serialized = syft_float._object2proto()

    assert isinstance(serialized, Float_PB)

    deserialized = Float._proto2object(proto=serialized)

    assert isinstance(deserialized, Float)
    assert deserialized.id == syft_float.id
    assert deserialized == syft_float


def test_send(client: sy.VirtualMachineClient) -> None:
    syft_float = Float(5)
    ptr = syft_float.send(client)
    # Check pointer type
    assert ptr.__class__.__name__ == "FloatPointer"

    # Check that we can get back the object
    res = ptr.get()
    assert res == syft_float



# stdlib
from dataclasses import dataclass
from enum import Enum


class CpuArchitectureTypes(Enum):
    x86_64 = 0


@dataclass
class CPU:
    num_cores: int
    num_threads_per_core: int
    clock_speed: float
    architecture: CpuArchitectureTypes
